"""

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-03-15
"""

import math
from typing import Dict, List, Tuple

from atligator.io_util import HiddenPrints, Colors
from atligator.mutation import Mutation
from atligator.prediction_matrix import BinderPredictionMatrix, BinderPredictionLine, BinderPredictionColumn

"""
Hydrophobicity of a residue.
Derived from: J. Kyte, R. F. Doolittle: A simple method for displaying the hydropathic character of a protein. 
Journal of Molecular Biology. Band 157, Nr. 1, 1982, S. 105–132, PMID 7108955. 
Recalculated from -10 (Arg: -4.5) to 10 (Ile: 4.5). 
"""
hp_index = {"ALA": 4.00, "ARG": -10.00, "ASN": -7.78, "ASP": -7.78, "CYS": 5.56,
            "GLN": -7.78, "GLU": -7.78, "GLY": -0.89, "HIS": -7.11, "ILE": 10.00,
            "LEU": 8.44, "LYS": -8.67, "MET": 4.22, "PHE": 6.22, "PRO": -3.56,
            "SER": -1.78, "THR": -1.56, "TRP": -2.00, "TYR": -2.89, "VAL": 9.33}


charge_index = {"ALA": 0.00, "ARG": 10.00, "ASN": 0.00, "ASP": -10.00, "CYS": 0.00,
                "GLN": 0.00, "GLU": -10.00, "GLY": 0.00, "HIS": 2.00, "ILE": 0.00,
                "LEU": 0.00, "LYS": 10.00, "MET": 0.00, "PHE": 0.00, "PRO": 0.00,
                "SER": 0.00, "THR": 0.00, "TRP": 0.00, "TYR": 0.00, "VAL": 0.00}


"""
Relative size of a residue
Derived from van der Waals radius. Recalculated from -10 (Gly: 48) to 10 (Trp: 163). 
"""
size_index = {"ALA": -6.70, "ARG": 7.39, "ASN": -1.65, "ASP": -2.52, "CYS": 3.39,
              "GLN": 1.48, "GLU": 0.61, "GLY": -10.00, "HIS": 2.17, "ILE": 3.22,
              "LEU": 3.22, "LYS": 5.13, "MET": 3.22, "PHE": 5.13, "PRO": -2.70,
              "SER": -5.65, "THR": -2.17, "TRP": 10.00, "TYR": 6.17, "VAL": -0.09}


def generate_clash_pmatrix(pmatrix: BinderPredictionMatrix,
                           parameters: Tuple[float, float, float, float, float]) -> BinderPredictionMatrix:
    max_scoresum = pmatrix.get_max_sumscore()
    pmatrix = pmatrix.normalize()
    with HiddenPrints():
        similarity_matrix = generate_similarity_scores(pmatrix, parameters, max_scoresum)
    return similarity_matrix.generate_derived_pmatrix()


class FeatureSet:
    """
    A set which combines one value for each feature (hydrophobicity, size and charge).
    """

    def __init__(self, hydrophobicity: float, size: float, charge: float, h_dev: float, s_dev: float, c_dev: float):
        self.hydrophobicity = hydrophobicity
        self.size = size
        self.charge = charge
        self.h_dev = h_dev
        self.s_dev = s_dev
        self.c_dev = c_dev

    def __str__(self) -> str:
        """converts a feature set into its string representation.
        :return: string serialization
        """
        hp = "{0:.2f}".format(self.hydrophobicity)
        h_dev = "{0:.2f}".format(self.h_dev)
        size = "{0:.2f}".format(self.size)
        s_dev = "{0:.2f}".format(self.s_dev)
        charge = "{0:.2f}".format(self.charge)
        c_dev = "{0:.2f}".format(self.c_dev)
        pm = " +-"
        string = "hp: " + hp + pm + h_dev + "   size: " + size + pm + s_dev + "   charge: " + charge + pm + c_dev
        return string

    def __repr__(self):
        return str(self)


class SimilarityMatrix:
    """
    A Matrix combining residue id (position) with all mutations.
    These mutations are combined by three letter code for amino acid (residue id), residue name, all mutations
    of this position with clash scores to average feature set and original residue (=delta) (generated
    by calculate_clash) and total quality
    for each position on the binding interface.
    """
    def __init__(self, positions: List,
                 max_scoresum: float,
                 base: float = 1.5,
                 delta_factor: float = 4.0,
                 n_delta_k: float = 5.0,
                 n_quality_k: float = 5.0):
        self.positions = positions
        self.n_delta_k = n_delta_k
        self.n_quality_k = n_quality_k
        self.base = base
        self.max_score = max_scoresum
        self.delta_factor = delta_factor

    def __iter__(self):
        for i in range(len(self.positions)):
            yield self.positions[i]

    def generate_derived_pmatrix(self) -> BinderPredictionMatrix:
        lines = []
        for position in self.positions:
            columns = []
            res_id = position[0]
            res_type = position[1]
            total_quality = position[3]
            for mutation in position[2]:
                mut_res_type = mutation[0]
                clash = mutation[1]
                delta = mutation[2]
                rating = self.rate(clash, total_quality, delta)
                columns.append(BinderPredictionColumn(mut_res_type, 1, rating))
            lines.append(BinderPredictionLine(res_id, res_type, columns))
        return BinderPredictionMatrix(lines)

    def rate(self, clash: float, total_quality: float, delta: float):
        n_delta = delta / (delta + self.n_delta_k)
        tq = total_quality
        clash_n_quality = self.base ** (-clash) * tq * self.max_score / (tq * self.max_score + self.n_quality_k)
        rating = n_delta/self.delta_factor + clash_n_quality
        return rating


def generate_similarity_scores(pmatrix: BinderPredictionMatrix, parameters: Tuple[float, float, float, float, float],
                               max_scoresum: float = 1.0) -> SimilarityMatrix:
    """
    Generates three mutations suggestions for each position in a binder prediction matrix.
    Takes P matrix as an imput where all mutations are read (similar to mutation_prediction.find_best_mutations)
    For each position an average value for all features (hydrophobicity, size and charge)
    gets calculated and compared with the values of all amino acids.
    The best three matches get returned
    :param pmatrix: a binder prediction matrix
    :param parameters: Parameters for weighing: clash_merge_base, delta_factor, n_delta, n_quality_k, std_dev_weight
    :param max_scoresum: The maximum sum of score (height of pmatrix bar) for one position
    :return: A SimilarityMatrix instance including clash scores
    """
    mutations = find_all_mutations(pmatrix)
    hydrophobicity_matrix = calculate_hydrophobicity(mutations)
    size_matrix = calculate_size(mutations)
    charge_matrix = calculate_charge(mutations)
    feature_matrix = []
    for i in range(len(size_matrix)):
        h = hydrophobicity_matrix[i][2]
        s = size_matrix[i][2]
        c = charge_matrix[i][2]
        h_dev = hydrophobicity_matrix[i][3]
        s_dev = size_matrix[i][3]
        c_dev = charge_matrix[i][3]
        features = FeatureSet(h, s, c, h_dev, s_dev, c_dev)
        residue_id = size_matrix[i][0]
        residue_name = size_matrix[i][1]
        total_quality = size_matrix[i][4]
        feature_matrix.append((residue_id, residue_name, features, total_quality))
    similarity_matrix = calculate_clash_matrix(feature_matrix, max_scoresum, parameters)
    return similarity_matrix


def find_all_mutations(pmatrix: BinderPredictionMatrix) -> List[List[Tuple[int, Mutation]]]:
    """
    Reads BinderPredictionMatrix object and returns all mutations with their quality.
    In contrast to find_best_mutations all mutations are considered - even exchanging to
    the same residue (= no mutation basically).
    These mutations are not sorted by their impact which is not necessary since their processed
    all the same way in feature calculation.
    :param pmatrix: Binder prediction matrix with mutations including score and certainty
    :return: List of all mutations which can be found in the pmatrix.
    """
    mutations = []
    for line in pmatrix.lines:
        position = []
        for col in line.columns:
            quality = col.score * int(col.certainty)
            mut = Mutation(line.residue_id, line.original_restype, col.residue_type)
            position.append((quality, mut))
        mutations.append(position)
    return mutations


def calculate_clash_matrix(feature_matrix: List[Tuple[int, str, FeatureSet, float]], max_scoresum: float,
                           par: Tuple[float, float, float, float, float]) -> SimilarityMatrix:
    """
    Calculates a matrix with all clash scores for all mutations in the p matrix,
    but just keeps the three mutations (at one position) with the lowest clash score.
    :param feature_matrix: Combines FeatureSets for all positions including resid, resname and total_quality
    :param max_scoresum: The maximum sum of score (height of pmatrix bar) for one residue.
    :param par: Parameters for weighing: clash_merge_base, delta_factor, n_delta, n_quality_k, std_dev_weight
    :return: A SimilarityMatrix object including clash scores
    """
    mutation_matrix = []
    for position in feature_matrix:
        res_id = position[0]
        res_name = position[1]
        total_quality = position[3]
        feature_set = position[2]
        lowest_clash = calculate_clash(feature_set, par[4], res_name)
        mutation_matrix.append((res_id, res_name, lowest_clash, total_quality))
    similarity_matrix = SimilarityMatrix(mutation_matrix, max_scoresum, par[0], par[1], par[2], par[3])
    return similarity_matrix


def calculate_clash_dep(features: FeatureSet, base: float = 1.5) -> Tuple[List, List, List]:
    """
    DEPRECATED
    Calculates a total clash for all amino acids regarding the current average feature values.
    Returns best three mutations.
    """
    best = [None, 30]
    second = [None, 30]
    third = [None, 30]
    for residue in hp_index:
        hp_clash = base ** - features.h_dev * abs(hp_index[residue] - features.hydrophobicity)
        size_clash = base ** - features.s_dev * abs(size_index[residue] - features.size)
        charge_clash = base ** - features.c_dev * abs(charge_index[residue] - features.charge)
        total_clash = hp_clash + size_clash + charge_clash
        if total_clash < best[1]:
            third = second
            second = best
            best = [residue, total_clash]
        elif total_clash < second[1]:
            third = second
            second = [residue, total_clash]
        elif total_clash < third[1]:
            third = [residue, total_clash]
    return best, second, third


def calculate_clash(features: FeatureSet, base: float, original_res: str = None) -> List[Tuple]:
    """
    Calculates a total clash for all amino acids regarding the current average feature values.
    The standard deviation allows to introduce a weight on different features:
        -> If standard deviation is high: Feature accuracy might be less important
        -> If standard deviation is low: "        "        might be very important
    base value for weighing can be adjusted (see parameters). If base == 1, no weighing is applied.

    If an original_res is defined, an additional clash is calculated:
    The clash between the current mutation residue and the original residue.
    This can serve as an indicator for a change in the features.
    ATTENTION: This calculation is also weighted, if base is not set to 1.

    Returns all legal (= stated in binder prediction matrix) mutations sorted by increasing clash score.
    """
    mutations = [('', 30.0, 0.0)]
    for residue in hp_index:
        hp_clash: float = base ** - features.h_dev * abs(hp_index[residue] - features.hydrophobicity)
        size_clash: float = base ** - features.s_dev * abs(size_index[residue] - features.size)
        charge_clash: float = base ** - features.c_dev * abs(charge_index[residue] - features.charge)
        total_clash: float = hp_clash + size_clash + charge_clash
        for mut in range(len(mutations)):
            if total_clash < mutations[mut][1]:
                clash_to_original = 0.0
                if original_res is not None:
                    hp_clash = base ** - features.h_dev * abs(hp_index[residue] - hp_index[original_res])
                    size_clash = base ** - features.s_dev * abs(size_index[residue] - size_index[original_res])
                    charge_clash = base ** - features.c_dev * abs(charge_index[residue] - charge_index[original_res])
                    clash_to_original = hp_clash + size_clash + charge_clash
                mutations.insert(mut, (residue, total_clash, clash_to_original))
                break
    mutations.pop()
    return mutations


def calculate_feature(index: Dict[str, float], mutations: List[List[Tuple[float, Mutation]]]) \
        -> List[Tuple[int, str, float, float, float]]:
    """
    Calculates the average score for each position.
    The type of score (feature: size, charge, etc.) is defined by the index which is passed.
    :param index: The index defining the feature type.
    :param mutations:
    :return:
    """
    feature_matrix = []
    counter = -1
    for position in mutations:
        counter += 1
        sum_ = 0
        total_quality = 0
        for mutation in position:
            feature_value = index[mutation[1].mutated_residue_type]
            quality = mutation[0]
            sum_ += quality * feature_value
            total_quality += quality
        if total_quality == 0:
            print("position", counter, "skipped due to missing data input.")
            continue
        average = sum_ / total_quality
        dev_sum = 0.0
        for mutation in position:
            feature_value = index[mutation[1].mutated_residue_type]
            quality = mutation[0]
            dev_sum += (feature_value - average)**2 * quality
        variance = dev_sum / total_quality
        std_dev = math.sqrt(variance)
        res_id = position[0][1].residue_id
        res_name = position[0][1].original_residue_type
        feature_matrix.append((res_id, res_name, average, std_dev, total_quality))
    return feature_matrix


def calculate_hydrophobicity(mutations):
    return calculate_feature(hp_index, mutations)


def calculate_size(mutations):
    return calculate_feature(size_index, mutations)


def calculate_charge(mutations):
    return calculate_feature(charge_index, mutations)


def print_output(similarity_matrix: SimilarityMatrix, args, pmatrix_max_sumscore, colors: Colors) -> None:
    """
    Prints similarity scores for each position.
    :param similarity_matrix: A matrix with residue id, residue name, mutations sorted by increasing clash
    and total quality for each position on the binding interface.
    :param args: Arguments given by argparser
    :param pmatrix_max_sumscore: The maximum sum of score (height of pmatrix bar) for one position
    :param colors: Colors instance for coloring
    :return: None
    """
    # Normalisation constant: k is the value of delta where it reaches 50 % in normalized state
    n_delta_k = args.n_delta_k
    n_quality_k = args.n_quality_k
    base = args.base
    max_ = pmatrix_max_sumscore
    print("\nMaximum score sum is:", max_)
    for position in similarity_matrix:
        # Intro: "ASN 45 - mutation:"
        if len(str(position[0])) == 3:
            print(position[1], position[0], "-           mutation", end=":   ")
        else:
            print(position[1], position[0], " -           mutation", end=":   ")
        # Header with residue type - also defines how many are displayed (counter):
        counter = 0
        for i in range(len(position[2])):
            if base ** (-position[2][i][1]) * position[3] * max_ / (position[3] * max_ + n_quality_k) < 0.05:
                break
            if position[2][i][0] == position[1]:
                print(colors.RED + position[2][i][0], " (o)  ", end=colors.WHITE)
            else:
                print(position[2][i][0], end="      ")
            counter += 1
            if counter == args.output_size_ex:
                break

        # clash scores - based on similarity of pmatrix data and static amino acid features - of best mutations
        # within range (atm controlled by normalized quality threshold)
        print("\n clash-score:  ┬─>to ori(del):   ", end="")

        if args.norm_del:
            for i in range(counter):
                delta = position[2][i][2]
                print("┼d>" + colors.DIM, "{0:.1f}".format(delta), " ", end=colors.WHITE)
            print(colors.WHITE + colors.BOLD + "\n clash-score:  │ └>norm_delta:   ", end="")
            for i in range(counter):
                delta = position[2][i][2]
                n_delta = delta / (delta + n_delta_k)
                print("└d>", "{0:.1f}".format(n_delta), " ", end="")
        else:
            colors.pc('bold')
            for i in range(counter):
                delta = position[2][i][2]
                print("┴d>", "{0:.1f}".format(delta), " ", end="")

        print(colors.WHITE + "\n               └─>to average :   ", end="")
        for i in range(counter):
            print("┬a>", "{0:.1f}".format(position[2][i][1]), " ", end="")

        # quality of mutation in pmatrix. Defined by score of clustering and certainty.
        # clash_quality is a mixed value by clash score and quality. defined by ' base^(-clash) * quality '
        if args.quality:
            print("\n  quality:  ", "{0:.2f}".format(position[3]), " - clash-qu:   ", end="")
            for i in range(counter):
                clash_quality = base ** (-position[2][i][1]) * position[3]
                print("├q>" + colors.DIM, "{0:.1f}".format(clash_quality), end="  " + colors.WHITE)

        # normalized quality. Takes into account that big differences at high qualities might not be worth punishing.
        # the basis is michaelis menten kinetics-like formula: 'value = (max_[=100%] * quality)/(quality + 50%-value)'
        # This leads to close to linear rating at small quality values and tiny differences at high quality values.
        norm_quality = 100 * position[3] * max_ / (position[3] * max_ + n_quality_k)
        norm_quality_str = "{0:.2f}".format(norm_quality) if norm_quality >= 10 else " " + "{0:.2f}".format(
            norm_quality)
        colors.pc('white')
        colors.pc('bold')
        print("\n n-quality: ", norm_quality_str, "% - clash-qu: ", sep="", end="  ")
        for i in range(counter):
            clash_quality = base ** (-position[2][i][1]) * norm_quality / 100
            print("└n>", "{0:.1f}".format(clash_quality), end="  ")
        print("\n" + colors.WHITE)


def rate_mutations(similarity_matrix: SimilarityMatrix, args, pmatrix_max_sumscore) -> \
        List[Tuple[str, str, str, float]]:
    """
    Sorts mutations in similarity matrix and returns list with list from best to worst.
    This list includes residue id (position), original residue type, mutated residue type and a rating.
    The rating is based on the formula: rating = norm_delta / delta_factor + clashed_quality.
    delta (the clash of mutation residue to original) is normalized at high values. This can be adjusted by
    n_delta_k (see argparse help).
    clashed_quality is calculated by base^-(clash to average) * norm_quality. base and quality normalization
    factor can be adjusted with argparse.
    :param similarity_matrix: A matrix with residue id, residue name, mutations sorted by increasing clash
    and total quality for each position on the binding interface.
    :param args: Arguments given by argparser
    :param pmatrix_max_sumscore: The maximum sum of score (height of pmatrix bar) for one position
    :return: None
    """
    sort = [("position", "original", "mutation", 0.0)]
    base = args.base
    # Normalisation constant: k is the value of delta where it reaches 50 % in normalized state
    n_delta_k = args.n_delta_k
    n_quality_k = args.n_quality_k
    # Correction factor for (n_quality*base^-clash + n_delta/FACTOR)
    delta_factor = args.delta_factor
    max_ = pmatrix_max_sumscore

    for position in similarity_matrix:
        # Header with residue type:
        for i in range(len(position[2])):
            if position[1] == position[2][i][0]:
                continue
            delta = position[2][i][2]
            n_delta = delta / (delta + n_delta_k)
            clash_n_quality = base ** (-position[2][i][1]) * position[3] * max_ / (position[3] * max_ + n_quality_k)
            for mut in range(len(sort)):
                # print(type(n_delta + clash_n_quality), type(sort[mut][2]))
                rating = n_delta / delta_factor + clash_n_quality
                if rating > sort[mut][3]:
                    sort.insert(mut, (position[0], position[1], position[2][i][0], rating))
                    break
    sort.pop()
    return sort


def show_mutations(similarity_matrix: SimilarityMatrix, args, pmatrix_max_sumscore, colors: Colors) -> None:
    """
    Showing the best 10 (or '-os') mutations in a human readable ranking.
    With the color flas '-c' all the doubled ones are dimmed after appearing lateron.
    :param similarity_matrix: A matrix with residue id, residue name, mutations sorted by increasing clash
    and total quality for each position on the binding interface.
    :param args: Arguments given by argparser
    :param pmatrix_max_sumscore: The maximum sum of score (height of pmatrix bar) for one position
    :param colors: Colors instance for coloring
    :return: None
    """
    sort = rate_mutations(similarity_matrix, args, pmatrix_max_sumscore)
    counter = 0
    print("rating original to mutation  (n_quality*base^-clash + n_delta/factor)")
    ids_out = []
    for i in sort:
        if i[0] in ids_out:
            colors.pc('dim')
        else:
            colors.pc('white')
            ids_out.append(i[0])
        if counter == args.output_size:
            break
        if len(str(counter + 1)) == 2:
            print(counter + 1, ": ", sep="", end="")
        else:
            print(counter + 1, ":  ", sep="", end="")
        print("  ", i[1], "", end="")
        dyn_space = ""
        for j in range(3 - len(str(i[0]))):
            dyn_space += " "
        print(i[0], dyn_space, sep="", end=" ")
        print(" to ", i[2], ":      {0:.2f}".format(i[3]), sep="")
        counter += 1


def safe_output(similarity_matrix: SimilarityMatrix, args, pmatrix_max_sumscore) -> None:
    """
    Saving a Mutation Set with the best 10 mutations (or changed by '-os' flag).
    :param similarity_matrix: A matrix with residue id, residue name, mutations sorted by increasing clash
    and total quality for each position on the binding interface.
    :param args: Arguments given by argparser
    :param pmatrix_max_sumscore: The maximum sum of score (height of pmatrix bar) for one position
    :return: None
    """
    sort = rate_mutations(similarity_matrix, args, pmatrix_max_sumscore)
    counter = 0

    for i in sort:
        if counter == args.output_size:
            break
        print(i[1], i[0], i[2], sep="", end=" ")
        counter += 1



"""This module contains utility functions for command line input and output.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-03-06
"""

import sys


def err(message: str) -> None:
    """Prints out an error message to stderr in a red color.
    :param message: The error message to print.
    """
    sys.stderr.write('\033[01;31m' + message + '\033[0;0m\n')


def reduce_float(floating_number: float, digits: int = 2):
    """
    Converts a floating number into a string and returns it with specified number of digits.
    :param floating_number: floating number to use
    :param digits: digits after the decimal point to display
    :return: string of floating number with 'digits' digits after decimal point
    """
    return ("{0:." + str(digits) + "f}").format(floating_number)


class Colors:
    """
    Class for coloring printed text. When initialized as colors allows to
    - adjust color/style of printed text by 'print(color.RED, "hello")'
    - change all prints by color.red (equal to 'print(color.RED, end="")')
    - change back to normal by color.white or color.WHITE (as a string in print)
    """
    LMAGENTA: str = '\033[95m'
    BLUE: str = '\033[94m'
    GREEN: str = '\033[32m'
    LGREEN: str = '\033[92m'
    YELLOW: str = '\033[93m'
    LRED: str = '\033[91m'
    GREY: str = '\033[90m'
    RED: str = '\033[31m'
    BLACK: str = '\033[30m'
    WHITE: str = '\033[0m'
    BOLD: str = '\033[1m'
    UNDERLINE: str = '\033[4m'
    DIM: str = '\033[2m'
    BLINK: str = '\033[5m'
    HIDDEN: str = '\033[8m'
    BACKGROUND: str = '\033[7m'

    def pc(self, item) -> None:
        """
        Prints the attribute of choice.
        :param item: attribute to print (e.g. a color)
        :return: None
        """
        print(getattr(self, item.upper()), end="")


class Blank(object):
    """
    Class as a blank placeholder for importing Colors class without changing colors or style.
    Use similar to: 'from atligator.io_util import Blank as Colors'  When coloring option is disabled.
    """
    def __getattribute__(self, item):
        return ''


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._original_stdout



"""The outcome of ATLIGATOR is the prediction of probable mutations on the binder site, increasing the binding
between binder and ligand. This module provides data structures and functions for mutation scoring. Scoring results
are kept in a so called binder prediction matrix.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-03-21
"""

from copy import deepcopy
from typing import List, Dict, Tuple, Set

from numpy import pi

from atligator.acomplex import LigandBinderComplex
from atligator.structure import AbstractResidue


class BinderPredictionColumn:
    """A column of a binder prediction line aggregates a residue type (indicating a mutation), a certainty value
    (number of atlas residues indicating the mutation, and the score value."""

    def __init__(self, residue_type: str, certainty: int, score: float):
        """Creates a new instance.
        :param residue_type: the candidate target type of a mutation
        :param certainty: the number of ligand residues from whose local atlases the mutation has been inferred
        :param score: relative probability of mutation success based on the overall atlas data
        """
        self.residue_type = residue_type
        self.certainty = certainty
        self.score = score

    def quality(self) -> float:
        """The quality of a column entry is defined as the product of score and certainty.
        :return: quality of this column
        """
        return self.score * float(self.certainty)

    def merge(self, other: 'BinderPredictionColumn') -> 'BinderPredictionColumn':
        """Merges this column with another column from another matrix by addding their scores and taking the maximum
        of their certainties.
        :param other: the column to merge with
        :return: a merged column
        """
        return BinderPredictionColumn(self.residue_type, max(self.certainty, other.certainty), self.score + other.score)


class BinderPredictionLine:
    """A line in a binder prediction aggregates the following data: the residue id of the involved binder residue,
    the original residue type, and a list of columns, each representing a suggested mutation."""

    def __init__(self, residue_id: int, original_restype: str, columns: List[BinderPredictionColumn]):
        """Creates a new instance.
        :param residue_id: the id of the binder residue candidate to mutation
        :param original_restype: the original residue type according to the scaffold
        :param columns: a list of columns, each predicting possible mutations for the context residue
        """
        self.residue_id = residue_id
        self.original_restype = original_restype
        self.columns = columns

    def merge(self, other: 'BinderPredictionLine') -> 'BinderPredictionLine':
        """Merges the columns of this matrix with corresponding columns of another, given matrix. As matching
        criteria for columns, we use the residue type.
        :param other: the line to merge with
        :return: a merged line containing merged columns
        """
        column_restypes = sorted(set([c.residue_type for c in self.columns + other.columns]))
        merged_columns: List[BinderPredictionColumn] = []
        for cr in column_restypes:
            if any([sc for sc in self.columns if sc.residue_type == cr]):
                sc = [sc for sc in self.columns if sc.residue_type == cr][0]
                if any([oc for oc in other.columns if oc.residue_type == cr]):
                    oc = [oc for oc in other.columns if oc.residue_type == cr][0]
                    merged_columns.append(sc.merge(oc))
                else:
                    merged_columns.append(deepcopy(sc))
            elif any([oc for oc in other.columns if oc.residue_type == cr]):
                oc = [oc for oc in other.columns if oc.residue_type == cr][0]
                merged_columns.append(deepcopy(oc))
        return BinderPredictionLine(self.residue_id, self.original_restype, merged_columns)


def read_binder_prediction_line(line: str) -> BinderPredictionLine:
    """Unparses a single line of a binder prediction matrix.
    :param line: the text serialization
    :return: a corresponding instance of BinderPredictionLine, including its columns
    """
    content = line.split()
    orig_restype = content[0]
    binder_res_id = int(content[1])
    columns: List[BinderPredictionColumn] = []
    for i in range(3, len(content), 3):
        residue_type = content[i]
        certainty = int(content[i+1])
        score = float(content[i+2])
        columns.append(BinderPredictionColumn(residue_type, certainty, score))
    return BinderPredictionLine(binder_res_id, orig_restype, columns)


class BinderPredictionMatrix:
    """A binder affinity matrix assigns to each designed residue of a ligand-binder complex a set of predicted
    scores for residue types."""

    def __init__(self, lines: List[BinderPredictionLine]):
        """Creates a new BinderPredictionMatrix from its lines.
        :param lines: The initial contents of the matrix.
        """
        self.lines = lines

    def __mul__(self, scalar: float) -> 'BinderPredictionMatrix':
        """Multiplies the scores of this matrix with a given scalar and returns the result
        :param scalar: the scalar value to multiply the score of each matrix cell with
        :return: a scaled prediction matrix
        """
        result_lines: List[BinderPredictionLine] = []
        for line in self.lines:
            result_columns: List[BinderPredictionColumn] = []
            for col in line.columns:
                result_columns.append(BinderPredictionColumn(col.residue_type, col.certainty, col.score * scalar))
            result_lines.append(BinderPredictionLine(line.residue_id, line.original_restype, result_columns))
        return BinderPredictionMatrix(result_lines)

    def get_max_sumscore(self, criterion: str = 'score') -> float:
        """Returns the max_sumscore, which denotes the inverse of the factor which is applied for normalization.
        :param criterion: the normalization criterion. Must be one of 'score', 'certainty', or 'quality'
        :return: the maximum sum of scores of all matrix lines
        """
        if criterion not in ['score', 'certainty', 'quality']:
            pass
        max_sumscore = 0.0
        for line in self.lines:
            sum_scores = 0.0
            for col in line.columns:
                if criterion == 'score':
                    sum_scores += col.score
                elif criterion == 'certainty':
                    sum_scores += col.certainty
                elif criterion == 'quality':
                    sum_scores += col.quality()
            max_sumscore = max(max_sumscore, sum_scores)
        if max_sumscore == 0.0:
            max_sumscore = 1.0
        return max_sumscore

    def get_total_sumscore(self, criterion: str = 'score') -> float:
        """Returns the sum of all score, certainty, or quality values contained in this matrix.
        :param criterion: the normalization criterion. Must be one of 'score', 'certainty', or 'quality'
        :return: the sum of values of all matrix lines
        """
        if criterion not in ['score', 'certainty', 'quality']:
            pass
        total_sumscore = 0.0
        for line in self.lines:
            for col in line.columns:
                if criterion == 'score':
                    total_sumscore += col.score
                elif criterion == 'certainty':
                    total_sumscore += col.certainty
                elif criterion == 'quality':
                    total_sumscore += col.quality()
        return total_sumscore

    def total_quality(self, ignore_identities: bool = False) -> float:
        """
        :return: the sum of quality values of all columns in all lines of this matrix
        """
        total_quality = 0.0
        for line in self.lines:
            for col in line.columns:
                if ignore_identities and line.original_restype == col.residue_type:
                    continue
                total_quality += col.quality()
        return total_quality

    def normalize(self, criterion: str = 'score', total: bool = False) -> 'BinderPredictionMatrix':
        """Rescales the matrix such that the maximum sum of scores, certainty, or quality contained in the columns of
        any line equals one. Keeps the context matrix untouched and returns the normalized variant.
        :param criterion: the normalization criterion. Must be one of 'score', 'certainty', 'quality'
        :param total: if True, the total score/certain/quality rather than the maximum is taken for scale
        :return: a normalized matrix
        """
        if criterion not in ['score', 'certainty', 'quality']:
            pass
        if total:
            return self * (1.0 / self.get_total_sumscore(criterion))
        else:
            return self * (1.0 / self.get_max_sumscore(criterion))

    def normalize_quality(self) -> 'BinderPredictionMatrix':
        """Shorthand for normalizing by quality.
        :return: a matrix normalized by quality
        """
        return self.normalize('quality')

    def sort_columns(self) -> 'BinderPredictionMatrix':
        """Sorts the columns of every line in the matrix by descending score. Returns the sorted matrix.
        :return: a column-sorted matrix
        """
        sorted_lines: List[BinderPredictionLine] = []
        for line in self.lines:
            sorted_columns = sorted(line.columns, key=(lambda col: col.quality()), reverse=True)
            sorted_lines.append(BinderPredictionLine(line.residue_id, line.original_restype, sorted_columns))
        return BinderPredictionMatrix(sorted_lines)

    def sort_lines(self) -> 'BinderPredictionMatrix':
        """Sorts the lines of this matrix by binder residue ID. Returns the sorted matrix.
        :return: a matrix sorted by lines
        """
        return BinderPredictionMatrix(sorted(self.lines, key=lambda l: int(l.residue_id)))

    def merge(self, other: 'BinderPredictionMatrix') -> 'BinderPredictionMatrix':
        """Merges the lines of this matrix with the lines of another given matrix and returns a merged matrix. The
        matching criterion is the residue id represented by each line.
        :param other: the other matrix to merge with
        :return: the merged matrix
        """
        line_resids = sorted(set([l.residue_id for l in self.lines + other.lines]))
        merged_lines: List[BinderPredictionLine] = []
        for lr in line_resids:
            if any([sl for sl in self.lines if sl.residue_id == lr]):
                sl = [sl for sl in self.lines if sl.residue_id == lr][0]
                if any([ol for ol in other.lines if ol.residue_id == lr]):
                    ol = [ol for ol in other.lines if ol.residue_id == lr][0]
                    merged_lines.append(sl.merge(ol))
                else:
                    merged_lines.append(deepcopy(sl))
            elif any([ol for ol in other.lines if ol.residue_id == lr]):
                ol = [ol for ol in other.lines if ol.residue_id == lr][0]
                merged_lines.append(deepcopy(ol))
        return BinderPredictionMatrix(merged_lines)

    def print_matrix(self, score_threshold: float = 0.001) -> None:
        """Prints out the columns of the matrix line by line
        :param score_threshold: score values below this threshold will not be printed
        """
        for line in self.lines:
            s = f"{line.original_restype} {line.residue_id} : "
            for col in line.columns:
                if col.score > score_threshold:
                    s += f"{col.residue_type} {col.certainty} {'{:10.6f}'.format(col.score)} "
            print(s)

    def save(self, filename, score_threshold: float = 0.001):
        """Saves this pmatrix into the specified file.
        :param filename: the desired serialization location
        :param score_threshold: columns indicating scores below this parameter are ignored for serialization
        """
        with open(filename, 'w') as fo:
            for line in self.lines:
                s = f"{line.original_restype} {line.residue_id} : "
                for col in line.columns:
                    if col.score > score_threshold:
                        s += f"{col.residue_type} {col.certainty} {'{:10.6f}'.format(col.score)} "
                fo.write(f"{s}\n")


def read_matrix(filename: str) -> BinderPredictionMatrix:
    """Converts a file containing a serialized binder prediction matrix into a runtime instance.
    :param filename: path of the file to open for deserialization
    :return: a parsed instance of BinderPredictionMatrix
    """
    lines: List[BinderPredictionLine] = []
    with open(filename, 'r') as fo:
        for line in fo:
            lines.append(read_binder_prediction_line(line))
    return BinderPredictionMatrix(lines)


def proximity_score(binder_res: AbstractResidue, aligned_res: AbstractResidue, score_threshold: float = 6.0,
                    distance_factor: float = 1.0, orient_factor: float = 6.0/pi, secor_factor: float = 6.0/pi,
                    score_exponent: float = 2.0) -> float:
    """Calculates the proximity score between a designed binder residue and an aligned residue. The higher this score
    is, the 'more proximate' the residues are. The score is in the interval [0; 1] and is indirectly proportional to a
    weighted sum of the distance, the angle between c_alpha->c_beta, and the angle between c_alpha->c_o.
    :param binder_res: the designed binder residue to compare
    :param aligned_res: the aligned residue to compare
    :param score_threshold: if the weighted sum of deviation values is above this value, the score is constantly 0.0
    :param distance_factor: the share of C_alpha-C_alpha distance in the scoring function [1 / Angstrom]
    :param orient_factor: the share of the angle between orientation vectors [1 / rad]
    :param secor_factor: the share of the secondary orientation angle in the weighted sum [1 / rad]
    :param score_exponent: before returning, the calculated score is raised with this value
    :return: a score in the interval between [0; 1] that describes the result of structural comparison
    """
    distance: float = (aligned_res.calpha() - binder_res.calpha()).norm()
    orient_angle: float = binder_res.orientation().angle(aligned_res.orientation())
    sec_orient_angle: float = binder_res.secondary_orientation().angle(aligned_res.secondary_orientation())
    wsum = distance_factor * distance + orient_factor * orient_angle + secor_factor * sec_orient_angle
    if wsum > score_threshold:
        return 0.0
    else:
        score = ((score_threshold - wsum) / score_threshold) ** score_exponent
        return score


def generate_prediction_matrix(acomplex: LigandBinderComplex, **score_args) -> BinderPredictionMatrix:
    """Evaluates the given ligand-binder complex and calculates binder prediction matrix by applying the scoring
    function for every pair of binder residue and aligned residue.
    :param acomplex: the ligand-binder complex to use as input for the calculations
    :param score_args: additional arguments that will be passed to the scoring function. See proximity_score
    :return: a non-normalized and unsorted binder prediction matrix generated
    """
    lines: List[BinderPredictionLine] = []
    n_aligned_res = len(acomplex.aligned_residues)
    for binres in acomplex.binder.residues:
        res_id = binres.residue_id
        columns_map: Dict[str, Tuple[str, Set[int], float]] = {}
        for ares in acomplex.aligned_residues:
            score = proximity_score(binres, ares, **score_args) / float(ares.size_factor) / float(n_aligned_res)
            if score > 0.0:
                if ares.binder_restype not in columns_map:
                    columns_map[ares.binder_restype] = (ares.binder_restype, set(), 0.0)
                col: Tuple[str, Set[int], float] = columns_map[ares.binder_restype]
                columns_map[ares.binder_restype] = (col[0], col[1].union({ares.ligand_res_position}), col[2]+score)
        columns: List[BinderPredictionColumn] = []
        for col in columns_map.values():
            columns.append(BinderPredictionColumn(col[0], len(col[1]), col[2]))
        lines.append(BinderPredictionLine(res_id, binres.original_residue_type, columns))
    return BinderPredictionMatrix(lines)



"""This module provides data structures and functions for prediction of mutations from pmatrices. These mutations can
be fed into external scoring programs after applying them to the scaffold structure.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-03-06
"""
import random
from copy import copy
from operator import itemgetter
from typing import List, Iterator, Tuple

from atligator.pdb_util import canonical_amino_acids
from atligator.prediction_matrix import BinderPredictionMatrix
from atligator.structure import CouplingInfo, DesignedBinder


class Mutation:
    """A mutation describes the exchange of a residue type at a defined location."""

    def __init__(self, residue_id: int, original_residue_type: str, mutated_residue_type: str, derived: bool = False,
                 representative: bool = False):
        """Creates a new instance.
        :param residue_id: the residue id within its chain
        :param original_residue_type: the original residue type to be changed by the mutation
        :param mutated_residue_type: the new residue type after mutation
        :param derived: whether this mutation has been derived from another mutation (e.g. by coupling)
        :param representative: whether this mutation is representative for a set of coupled mutations
        """
        self.residue_id = residue_id
        self.original_residue_type = original_residue_type
        self.mutated_residue_type = mutated_residue_type
        self.derived = derived
        self.representative = representative

    def can_combine(self, other: 'Mutation') -> bool:
        """This mutation can be combined with another mutation if both instances refer to different residue ids.
        :param other: the other mutation to check for combinability
        :return: whether the combination is valid
        """
        return not self.residue_id == other.residue_id

    def is_identity(self) -> bool:
        """
        :return: whether this mutation is identical, i.e., has no effect.
        """
        return self.original_residue_type == self.mutated_residue_type

    def get_coupled_mutations(self, coupling_info: List[CouplingInfo], designed_binder: DesignedBinder) \
            -> List['Mutation']:
        """
        Determines a list of mutations coupled with this mutation based on provided coupling information.
        :param coupling_info: describes which residues are coupled
        :param designed_binder: the region of the acomplex from which to take original residue types
        :return: a list of coupled mutations, including the context mutation itself
        """
        if coupling_info is not None:
            for ci in coupling_info:
                if len(ci[1]) < 2:
                    continue
                for resid in ci[1]:
                    if int(resid) == self.residue_id:
                        representant = ci[1][0]
                        original_restypes = ",".join(sorted(set([dr.original_residue_type for dr
                                                                 in designed_binder.residues
                                                                 if str(dr.residue_id) in ci[1]])))
                        return [Mutation(int(coupled_resid), original_restypes, self.mutated_residue_type,
                                         derived=(coupled_resid != representant),
                                         representative=(coupled_resid == representant))
                                for coupled_resid in ci[1]]
        return [self]

    def __str__(self) -> str:
        """converts a mutation into its string representation.
        :return: string serialization
        """
        return f"{self.original_residue_type}{self.residue_id}{'+' if self.representative else ''}" \
               f"{'/' if self.derived else ''}{self.mutated_residue_type}"

    def __eq__(self, other: 'Mutation') -> bool:
        """Considers this mutation's residue id, original type, and mutated type for comparison.
        :param other: the mutation to compare with
        :return: whether self and other are equal
        """
        return self.residue_id == other.residue_id and \
               self.mutated_residue_type == other.mutated_residue_type and \
               self.original_residue_type == other.original_residue_type and \
               self.derived == other.derived and self.representative == other.representative

    def equiv(self, other: 'Mutation') -> bool:
        """
        Performs an equality check ignoring the original residue type as well as the derived/representative flags.
        :param other: the mutation to compare with
        :return: wehther self and other are equivalent
        """
        return self.residue_id == other.residue_id and \
               self.mutated_residue_type == other.mutated_residue_type

    def __hash__(self) -> int:
        return hash(self.residue_id) + hash(self.original_residue_type) + hash(self.mutated_residue_type)


def read_mutation(word: str) -> Mutation:
    """Creates a mutation from its text serialization.
    :param word: the string to convert
    :return: a parsed runtime instance
    """
    residue_id = ""
    original_residue_type = ""
    mutated_residue_type = ""
    n_digits = 0
    derived = False
    coupled = False
    for c in word:
        if c.isdigit():
            n_digits += 1
            residue_id += c
        elif c == '+':
            coupled = True
        elif c == '/':
            derived = True
        elif n_digits < 1:
            original_residue_type += c
        else:
            mutated_residue_type += c
    return Mutation(int(residue_id), original_residue_type, mutated_residue_type,
                    representative=coupled, derived=derived)


class MutationSet:
    """A mutation set combines several mutations that do not contradict each other."""

    def __init__(self, mutations: List[Mutation]):
        """Creates a new mutation set.
        :param mutations: the initial content of the mutations list
        """
        self.mutations = mutations

    def combine(self, mutation: Mutation) -> 'MutationSet':
        """Combines this mutation set with a given mutation.
        :param mutation: the mutation to be combined with this mutation set
        :return: a combined mutation set, or None if the combination is illegal
        """
        if all(m.can_combine(mutation) for m in self.mutations):
            return MutationSet(self.mutations + [mutation])

    def crossbreed(self, other: 'MutationSet') -> 'MutationSet':
        """Combines the mutations of this mutation set with those of another one and returns a cross-breeded instance.
        In the case of conflicts, this mutation set is dominant, i.e., the mutated residue type of this mutation and
        the original residue type of the other mutation are accepted.
        :param other: the other mutation set to crossbreed with
        :return: a combined mutation set
        """
        muts = [copy(sm) for sm in self.mutations]
        if other is not None:
            for om in other.mutations:
                conflicting_muts = [sm for sm in self.mutations if sm.residue_id == om.residue_id]
                if len(conflicting_muts) == 0:
                    muts.append(copy(om))
                else:
                    for cm in conflicting_muts:
                        cm.original_residue_type = om.original_residue_type
        return MutationSet(muts)

    def __str__(self) -> str:
        """
        :return: a text serialization of this mutation set
        """
        result = ""
        for mut in sorted(self.mutations, key=lambda m: m.residue_id):
            if not mut.derived:
                result += f"{mut} "
        result = result[0:-1]
        return result

    def __eq__(self, other: 'MutationSet') -> bool:
        """Considers this mutation set's mutations for comparison.
        :param other: the mutation to compare with
        :return: whether self and other are equal
        """
        return len(self.mutations) == len(other.mutations) and all(mut in other.mutations for mut in self.mutations)

    def equiv(self, other: 'MutationSet') -> bool:
        """Considers this mutation set's mutations for equivalence. In contrast to equality comparison, the original
        residue types as well as the derived/representative flags of mutations are ignored.
        :param other: the mutation to compare with
        :return: whether self and other are equivalent
        """
        return len(self.mutations) == len(other.mutations) and (len(self.mutations) == 0 or all(
            any(m for m in other.mutations if m.equiv(mut)) for mut in self.mutations))

    def __hash__(self):
        return sum([hash(mut) for mut in self.mutations])

    def remove_mutation(self, res_id: int):
        for i_n, i in enumerate(self.mutations):
            if i.residue_id == res_id:
                del self.mutations[i_n]
                break


def read_mutation_set(line: str) -> MutationSet:
    """Recovers a mutation set from its string serialization.
    :param line: the string serialization
    :return: a mutation set containing the mutations parsed from this line
    """
    muts: List[Mutation] = []
    for word in line.split():
        muts.append(read_mutation(word))
    return MutationSet(muts)


def read_mutation_sets(filename: str) -> Iterator[MutationSet]:
    """Reads in a file that contains a collection of mutation sets, each represented by a line of text
    :param filename: the path of the file to open
    :return: a generator of mutation sets parsed from the file
    """
    with open(filename, 'r') as fo:
        for line in fo:
            yield read_mutation_set(line)


def recombine(mutation_sets: List[MutationSet], mutations: List[Mutation]) -> Iterator[MutationSet]:
    """Recombines a list of mutation sets with a list of base mutations. The resulting mutation set contains up to
    len(mutation_sets)*len(mutations) elements. All elements are valid mutations.
    :param mutation_sets: the mutation sets to recombine
    :param mutations: the mutations to apply to each mutation set if legal
    :return: a generator for recombined mutation sets. It does not yield the original contents of mutation_sets.
    """
    for ms in mutation_sets:
        for mut in mutations:
            combined_ms = ms.combine(mut)
            if combined_ms is not None:
                yield combined_ms


def find_best_mutations(pmatrix: BinderPredictionMatrix, n_mutations: int, allow_identities: bool = False) \
        -> List[Mutation]:
    """Finds the best n mutations suggested by a prediction matrix based on the quality of the mutation that can be
    deduced by each column. Only non-identical mutations are considered.
    :param pmatrix: the prediction matrix to analyze
    :param n_mutations: the number of mutations to find
    :param allow_identities: if true, pseudo-mutations that do not change resdue type are included in the result
    :return: a list of mutations (maximum size n_mutations) having best quality
    """
    mutations: List[Tuple[float, Mutation]] = []
    for line in pmatrix.lines:
        for col in line.columns:
            mut = Mutation(line.residue_id, line.original_restype, col.residue_type)
            if allow_identities or not mut.is_identity():
                mutations.append((col.quality(), mut))
    sorted_mutations = sorted(mutations, key=itemgetter(0), reverse=True)
    return list([e[1] for e in sorted_mutations])[0:n_mutations]


def generate_mutations(source: BinderPredictionMatrix or MutationSet, base_mutations: int, n_combinations: int,
                       allow_identities: bool = True) -> List[MutationSet]:
    """Generates a collection of mutation sets based on a binder prediction matrix.
    :param source: the prediction matrix to use as basis for mutations
    :param base_mutations: the number of atomic mutations to consider
    :param n_combinations: the maximum length of mutation sets created by recombination of base mutations
    :param allow_identities: if true, pseudo-mutations that do not change resdue type are included in the result
    :return: a list of mutation sets generated from this matrix and the specified parameters
    """
    if type(source) == BinderPredictionMatrix:
        mutations = find_best_mutations(source, base_mutations, allow_identities)
    else:
        mutations = list(mutation for mutation in source.mutations)[0:base_mutations]
    mutation_sets = list([MutationSet([m]) for m in mutations])
    for i in range(1, n_combinations):
        for ms in recombine(list(mutation_sets), mutations):
            if ms not in mutation_sets:
                mutation_sets.append(ms)
    return mutation_sets


# TODO Deprecated:
def select_random_matrix_mutation(pmatrix: BinderPredictionMatrix, allow_identities: bool = False) -> Mutation:
    """Selects a random mutation from this matrix. The probability for taking a specific random mutation is based on the
    quality factor of the respective column.
    :param pmatrix: taken as a basis for mutation selection
    :param allow_identities: if True, the selected mutation may be identical (e.g. SER123SER)
    :return: an instance of Mutation representing the random pick
    """

    def _random_pick() -> Mutation:
        """Utility function that performs a random pick in a pmatrix
        :return: the random generated mutation
        """
        quality_slice: float = random.uniform(0.0, pmatrix.total_quality())
        curr_qual: float = 0.0
        for pline in pmatrix.lines:
            for pcolumn in pline.columns:
                curr_qual += pcolumn.quality()
                if curr_qual > quality_slice:
                    return Mutation(int(pline.residue_id), pline.original_restype, pcolumn.residue_type)

    while True:
        mut = _random_pick()
        if mut is not None and (allow_identities or not mut.is_identity()):
            return mut


def select_random_mutation(pmatrix: BinderPredictionMatrix, allow_identities: bool = False,
                           allow_all: bool = False, random_mutations: bool = False, chance: float = 0.2) -> Mutation:
    """Selects a random mutation in either of two ways:
    By totally random selection of location and amino acid type OR by selection with pmatix probabilities.
    For totally random picks the probability for taking a specific random mutation is the same for
    every wild card and every mutation (except from PRO, CYS and GLY).
    The probability for taking a specific random mutation from the pmatrix is based on the
    quality factor of the respective column.
    :param pmatrix: Defines binder residues. Is also used for mutation selection.
    :param allow_identities: If True, the selected mutation may be identical (e.g. SER123SER)
    :param allow_all: Whether to allow mutations to all canonical amino acids. If False, Gly, Pro & Cys are not allowed.
    :param random_mutations: Whether to apply random mutations independent from pmatrix.
    :param chance: Chance of picking a totally random mutation
    :return: an instance of Mutation representing the random pick
    """

    def _random_pick() -> Mutation:
        """Utility function that performs a random pick
        :return: the random generated mutation
        """
        pline = random.choice(pmatrix.lines)
        return Mutation(int(pline.residue_id), pline.original_restype, random.choice(amino_acids))

    def _random_matrix_pick() -> Mutation:
        """Utility function that performs a random pick in a pmatrix
        :return: the random generated mutation
        """
        quality_slice: float = random.uniform(0.0, pmatrix.total_quality())
        curr_qual: float = 0.0
        for pline in pmatrix.lines:
            for pcolumn in pline.columns:
                curr_qual += pcolumn.quality()
                if curr_qual > quality_slice:
                    return Mutation(int(pline.residue_id), pline.original_restype, pcolumn.residue_type)

    if allow_all:
        amino_acids = canonical_amino_acids[:]
    else:
        amino_acids = canonical_amino_acids[:]
        amino_acids.remove("CYS")
        amino_acids.remove("PRO")
        amino_acids.remove("GLY")
    while True:
        if random_mutations and random.random() < chance:
            mut = _random_pick()
        else:
            mut = _random_matrix_pick()
        if mut is not None and (allow_identities or not mut.is_identity()):
            return mut



"""This module contains the implementation of a genetic mutation strategy that utilizes an atlas for calculating a
probability matrix for mutations dynamically, and a scorer for assessing the fitness of individuals.

The algorithm takes as input a list of suggested mutations. It applies these mutations to the scaffold structure in
isolation and determines whether the binding has improved based on the total energy calculated. Energy-increasing
mutations are then recombined for a given number of generations, where the individuals having the highest score
are given priority.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-03-28
"""

import logging as log
import os
import pickle
import random
from concurrent.futures import ThreadPoolExecutor
from copy import copy
from datetime import timedelta
from timeit import default_timer as timer
from typing import List, Tuple, Dict, Callable

from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Residue import Residue
from Bio.PDB.Structure import Structure

from atligator.acomplex import LigandBinderComplex
from atligator.acomplex import generate_complex, ComplexDescriptor
from atligator.atlas import Atlas
from atligator.base_scorer import BaseScorer
from atligator.deep_atlas import generate_deep_prediction_matrix, DeepAtlas
from atligator.mutation import MutationSet, select_random_mutation, Mutation
from atligator.pdb_util import mutate_residue, aa_3to1_conv
from atligator.prediction_matrix import generate_prediction_matrix, BinderPredictionMatrix, read_matrix
from atligator.similarity_scoring import generate_clash_pmatrix


class Individual:
    """Instances of this class represent invididuals of the population managed by a genetic mutator."""
    template_structure: str  # path to the PDB file this individual was derived from
    epoch: int  # the epoch in which the individual was created
    generation: int  # the generation within the epoch in which the individual was created
    iid: int  # a unique identifier of the individual that is usually managed by the mutator
    fitness: float  # The fitness calculated by the scorer, reflecting the fitness of the individual
    score_terms: Dict[str, float]  # score terms underlying the fitness
    optimized_structure: str  # PDB file that contains the optimized structure
    progenitor: 'Individual'  # the progenitor individual (belonging to the epoch)
    mother: 'Individual'  # mother (should exist for every non-scaffold individual)
    father: 'Individual'  # father (should exist for every non-scaffold and non-orphan individual)

    def __init__(self, mset: MutationSet):
        """Creates a new individual based on its mutation set
        :param mset: the set of mutations this individual reflects
        """
        self.mset = mset
        self.template_structure = None
        self.epoch = -1
        self.generation = -1
        self.iid = -1
        self.fitness = None
        self.score_terms = None
        self.optimized_structure = None
        self.progenitor = None
        self.mother = None
        self.father = None

    def crossbreed(self, father: 'Individual') -> 'Individual':
        """Combines two individuals by crossbreeding their mutation sets. Assumes the current individual as mother
        :param father: the individual to breed with
        :return: a cross-breeded individual
        """
        child = Individual(self.mset.crossbreed(father.mset))
        child.mother = self
        child.father = father
        return child

    def absolute_mset(self) -> MutationSet:
        """returns a mutation set that describes the mutations assigned to this individual, plus non-conflicting
        predecessor mutations from the progenitor individual, if any.
        :return: the absolute mutation set of this individual
        """
        if self.progenitor is None:
            return self.mset
        else:
            return self.mset.crossbreed(self.progenitor.absolute_mset())

    def __str__(self) -> str:
        return '{:10.4f} '.format(self.fitness) + '{:3} '.format(self.epoch) + \
               '{:3} '.format(self.generation) + '{:5} '.format(self.iid) + f"[{self.absolute_mset()}]" + " (" + \
               ' '.join([str(k) + ':' + '{:10.4f} '.format(v).strip() for k, v in self.score_terms.items()]) + ")"


class GeneticMutator:
    """A genetic mutator controls the generation of individuals over multiple epochs and generations. This class
    summarizes the data necessary for management and offers high-level controlling operations.
    """
    current_iid: int  # the id of the last individual generated
    generation: int  # keeps track of the current generation within the epoch
    population: List[Individual]  # the overall population
    msets: List[str]  # the overall population's msets in string representation for faster searching
    progens_by_epoch: List[Individual]  # list of progenitors; the index corresponds to the epoch in which used
    acomplex_by_epoch: List[LigandBinderComplex]  # acomplexes used for pmatrix derivation in every epoch
    pmatrix_by_epoch: List[BinderPredictionMatrix]  # pmatrices used for mutation prediction in every epoch
    previous_top_n: List[Individual]  # stores the best n_top individuals of the previous generation.

    def __init__(self, atlas: Atlas, deep_atlas: DeepAtlas, descriptor: ComplexDescriptor, n_top: int,
                 clustering_factor: float, clash_factor: float, deep_factor: float, random_mutations: bool,
                 random_mutation_chance: float, max_siblings: int, n_clones: int, delta_factor: float,
                 clash_merge_base: float, n_delta: float, std_dev_weight: float, n_quality_k: float):
        """Creates a new genetic mutator based on an atlas and user-defined input.
        :param atlas: the atlas to use for the creation of binder prediction matrices in every epoch
        :param deep_atlas: optional; adds deep learning facilities to pmatrixd calculation
        :param descriptor: user-defined input (ligand residue types, mutable host residues, couplings)
        :param n_top: the number of individuals to print after every epoch; a generation is to be terminated if this
        list does not change
        :param clustering_factor: probability of distance clustering based mutation selection
        :param clash_factor: probability to choose clash factoring for the selection of mutations
        :param deep_factor: probability to involve a deep learning strategy for mutation selection
        :param random_mutations: whether to apply random mutations.
        :param random_mutation_chance: The chance of random mutations (over ones based on pmatrix).
        :param max_siblings: the maximum number of individuals with identical mutation sets that may coexist per epoch
        :param n_clones: the number of clones automatically generated for new individuals (identical mutation sets).
        Original is also counted. -> n_clones 1 means individuals are not cloned!
        :param delta_factor: Correction factor for merging clash/quality with delta within calculation of mutation
        ranking: 'n_quality*base^-clash + n_delta/FACTOR'.
        :param clash_merge_base: Base of clash punishment for merging with normalized quality: n_quality*BASE^-clash.
        :param n_delta: Normalization constant for delta. Equals to delta value of 50%% in normalized state.
        :param std_dev_weight: Base in weighing the standard deviation: BASE^-std_dev * clash.
        :param n_quality_k: Normalization constant for quality. Equals to quality value of 50%% in normalized state.
        """
        self.atlas = atlas
        self.deep_atlas = deep_atlas
        self.descriptor = descriptor
        self.n_top = n_top
        self.clustering_factor = clustering_factor
        self.clash_factor = clash_factor
        self.deep_factor = deep_factor
        self.random_mutations = random_mutations
        self.random_mutation_chance = random_mutation_chance
        self.max_siblings = max_siblings
        self.n_clones = n_clones
        self.clash_params = clash_merge_base, delta_factor, n_delta, n_quality_k, std_dev_weight
        self.current_iid = 0
        self.generation = 0
        self.population = []
        self.msets = []
        self.progens_by_epoch = []
        self.acomplex_by_epoch = []
        self.pmatrix_by_epoch = []
        self.previous_top_n = []

    def current_epoch(self) -> int:
        """:return: the current epoch"""
        if len(self.progens_by_epoch) > 0:
            return len(self.progens_by_epoch)
        else:
            return 0

    def current_progenitor(self) -> Individual:
        """:return: the progenitor structure used in the current epoch"""
        if len(self.progens_by_epoch) > 0:
            return self.progens_by_epoch[-1]

    def current_acomplex(self) -> LigandBinderComplex:
        """:return: the acomplex used in the current epoch"""
        if len(self.acomplex_by_epoch) > 0:
            return self.acomplex_by_epoch[-1]

    def current_pmatrix(self) -> BinderPredictionMatrix:
        """:return: the pmatrix used in the current epoch"""
        if len(self.pmatrix_by_epoch) > 0:
            return self.pmatrix_by_epoch[-1]

    def initialize_deep_atlas(self):
        """Initializes the deep atlas of this mutator using the default parametration."""
        self.deep_atlas.train(self.atlas)

    def start_epoch(self, default_progenitor_path: str = None) -> None:
        """Starts a new epoch. Sets a new progenitor based upon the list of globally fittest individual such that the
        new progenitor does not have a mutation set indentical to the one of any preceding progenitor. It is assumed
        that update_acomplex and update_pmatrix are the first two operations to be called after starting a new epoch.
        :param default_progenitor_path: the structure to serve as progenitor in the new epoch in case there is no other
        progenitor available from the list of fittest individuals
        structure of all previous epochs is selected automatically
        :return the number of the current epoch
        """
        self.generation = 0
        fittest = self.sorted_population(False)
        progenitor = None
        for f in fittest:  # TODO hash?
            if not any(progen for progen in self.progens_by_epoch if progen.absolute_mset().equiv(f.absolute_mset())):
                progenitor = f
                break
        if progenitor is None:
            progenitor = Individual(MutationSet([]))
            progenitor.template_structure = default_progenitor_path
            progenitor.optimized_structure = default_progenitor_path
        progenitor.progenitor = self.current_progenitor()
        self.progens_by_epoch.append(progenitor)

    def update_acomplex(self, acomplexfile: str) -> LigandBinderComplex:
        """To be called explicitly after starting any epoch. Generates a new acomplex based on the current progenitor,
        saves it under the specified path, and returns it.
        :param acomplexfile: storage location of the newly created acomplex
        :return: the generated acomplex
        """
        progen_structure: Structure = PDBParser(QUIET=True).get_structure(self.current_progenitor().optimized_structure,
                                                                          self.current_progenitor().optimized_structure)
        acomplex = generate_complex(self.atlas, progen_structure, self.descriptor)
        with open(acomplexfile, 'wb') as fo:
            pickle.dump(acomplex, fo)
        self.acomplex_by_epoch.append(acomplex)
        return acomplex

    def update_pmatrix(self, pmatrixfile: str, **score_args) -> BinderPredictionMatrix:
        """To be called explicitly after updating the acomplex of any epoch. Generates a new pmatrix based on the
        current acomplex, saves it under the specified paths, and returns it.
        :param pmatrixfile: storage location of the newly generated acomplex
        :param score_args: will be passed to the scoring function
        :return: the generated pmatrix
        """
        pmatrix = generate_prediction_matrix(self.current_acomplex(), **score_args)
        pmatrix = pmatrix.normalize(criterion='quality', total=True) * self.clustering_factor
        if self.clash_factor > 0.0:
            clash_pmatrix = generate_clash_pmatrix(pmatrix, self.clash_params) \
                .normalize(criterion='quality', total=True)
            pmatrix = pmatrix.merge(clash_pmatrix * self.clash_factor)
        if self.deep_factor > 0.0:
            deep_pmatrix = generate_deep_prediction_matrix(self.current_acomplex(), self.deep_atlas) \
                .normalize(criterion='quality', total=True)
            pmatrix = pmatrix.merge(deep_pmatrix * self.deep_factor)
        pmatrix = pmatrix.normalize().sort_columns().sort_lines()
        if pmatrix.total_quality(True) > 0.0:
            pmatrix.save(pmatrixfile)
            self.pmatrix_by_epoch.append(pmatrix)
            return pmatrix

    def sorted_population(self, restrict_to_current_epoch: bool, max_elements: int = -1) -> List[Individual]:
        """Identifies the list of fittest individuals
        :param restrict_to_current_epoch: if True, individuals belonging to epochs different from the current are
        ignored
        :param max_elements: maximum length of the returned list
        :return: the fittest individuals with ascending score values
        """
        ce = self.current_epoch()
        spop = sorted([i for i in self.population if
                       (not restrict_to_current_epoch or i.epoch == ce) and (i.fitness is not None)],
                      key=lambda i: i.fitness)
        if max_elements > 0:
            return spop[:max_elements]
        else:
            return spop

    def start_generation(self) -> int:
        """Starts a new generation within the current epoch.
        :return: the generation number.
        """
        self.previous_top_n = self.sorted_population(True, self.n_top)
        self.generation += 1
        return self.generation

    def apply_random_mutations(self, ind: Individual, random_repeat: float) -> int:
        """Applies random mutations to the given individual until it fulfills sibling constraints and according to a
        given random repeat probability
        :param ind: the individual to apply random mutations to
        :param random_repeat: with this probability, random mutations are even applied if not necessary
        :return: the randomly mutated individual
        """
        n_rand = 0
        hashed_mset = mset_to_hash(ind.absolute_mset())
        while random.random() < random_repeat or \
                len([i for i in self.msets if i == hashed_mset]) >= self.max_siblings + 1 - self.n_clones:
            if len(ind.mset.mutations) > 0 and random.random() < 0.33:
                ran_mut: Mutation = random.choice(ind.mset.mutations)
                ran_cmuts = ran_mut.get_coupled_mutations(self.descriptor.coupling_info, self.current_acomplex().binder)
                for mut in ran_cmuts:
                    ind.mset.remove_mutation(mut.residue_id)
            else:
                rmut = select_random_mutation(self.current_pmatrix(), random_mutations=True,
                                              chance=self.random_mutation_chance)
                cmuts = rmut.get_coupled_mutations(self.descriptor.coupling_info, self.current_acomplex().binder)
                ind.mset = ind.mset.crossbreed(MutationSet(cmuts))
            n_rand += 1
            hashed_mset = mset_to_hash(ind.absolute_mset())
        return n_rand

    def create_progenitor_individual(self):
        """Generates a new individual that has exactly the properties of the progenitor.
        :return: the individual generated
        """
        ind = Individual(MutationSet([]))
        ind.progenitor = self.current_progenitor()
        ind.epoch = self.current_epoch()
        ind.generation = self.generation
        self.apply_random_mutations(ind, 0.0)
        return ind

    def create_orphan_individual(self, mother: Individual, random_repeat: float = 0.75) \
            -> Tuple[int, Individual]:
        """Generates a new individual based on an origin individual and a couple of random mutations.
        :param mother: the precursor individual
        :param random_repeat: probability of random mutations; repeatedly applied
        :return: a tuple consisting of the number of random mutations applied and the individual generated
        """
        if mother is None:
            mother = self.create_progenitor_individual()
        orphan = Individual(MutationSet(mother.mset.mutations[:]))
        orphan.mother = mother
        orphan.progenitor = self.current_progenitor()
        orphan.epoch = self.current_epoch()
        orphan.generation = self.generation
        n_rand = self.apply_random_mutations(orphan, random_repeat)
        return n_rand, orphan

    def create_breeded_individual(self, mother: Individual, father: Individual, random_repeat: float = 0.25) \
            -> Tuple[int, Individual]:
        """Generates a new individual by crossbreeding a given mother and father individual. If necessary or with a
        specified probability, random mutations are applied in addition. It is ensured that no identical individual
        is already managed by this mutator.
        :param mother: the mother individual; will be given priority in all mutation conflicts
        :param father: the father individual
        :param random_repeat: probability of random mutations; repeatedly applied
        :return: a tuple consisting of the number of random mutations applied and the individual breeded
        """
        child = mother.crossbreed(father)
        child.progenitor = self.current_progenitor()
        child.epoch = self.current_epoch()
        child.generation = self.generation
        n_rand = self.apply_random_mutations(child, random_repeat)
        return n_rand, child

    def has_improved_during_generation(self) -> bool:
        """:return: whether the top n individuals have not improved since finishing the last generation
        """
        return not self.previous_top_n == self.sorted_population(True)[:self.n_top]

    def insert_individual(self, ind: Individual) -> None:
        """Inserts an individual, assigns a new iid to it, and increments the current iid counter.
        Also stores a short version of total mutations set called "hash" in the mutator for faster checks.
        :param ind: the individual to insert
        """
        ind.iid = self.current_iid
        self.msets.append(mset_to_hash(ind.absolute_mset()))
        self.population.append(ind)
        self.current_iid += 1


def muts_to_hash(mutations: List[Tuple[int, str]]):
    hash = "" if len(mutations) >= 1 else "[]"
    mutation_list = []
    for mut in mutations:
        mutation_list.append((mut[0],
                              aa_3to1_conv[mut[1]]))
    mutation_list = sorted(mutation_list)
    for mut in mutation_list:
        hash += f"{mut[0]}{mut[1]}"
    return hash


def mset_to_hash(mset: MutationSet) -> str:
    hash = "" if len(mset.mutations) >= 1 else "[]"
    mutation_list = []
    for mut in mset.mutations:
        mutation_list.append((mut.residue_id,
                              aa_3to1_conv[mut.mutated_residue_type]))
    mutation_list = sorted(mutation_list)
    for mut in mutation_list:
        hash += f"{mut[0]}{mut[1]}"
    return hash


def mutate_mset(structure: Structure, chain_id: str, mset: MutationSet) -> None:
    """Applies one or more point mutations, described by a mutation set, to a given structure.
    :param structure: original structure not containing the mutation
    :param chain_id: the id of the chain where to apply the mutations
    :param mset: the mutations to apply
    """
    for mutation in mset.mutations:
        residue = structure[0][chain_id][int(mutation.residue_id)]
        mutate_residue(residue, mutation.mutated_residue_type)


def mutate_scaffold(scaffold_path: str, descriptor: ComplexDescriptor) -> None:
    """Applies the mutations described by the ligand info of a specified descriptor to a scaffold file. Saves the result
    under the same path.
    :param scaffold_path: a PDB file to which all ligand mutations are applied
    :param descriptor: its ligand info contains the ligand mutations to apply
    """
    parser = PDBParser(QUIET=True)
    scaffold_structure: Structure = parser.get_structure(scaffold_path, scaffold_path)
    for chain_id, residue_id, new_residue_type in descriptor.ligand_info:
        residue: Residue = scaffold_structure[0][chain_id][int(residue_id)]
        mutate_residue(residue, new_residue_type)
    io = PDBIO()
    io.set_structure(scaffold_structure)
    io.save(scaffold_path)


def genetic_mutate(scorer: BaseScorer, atlas: Atlas, deep_atlas: DeepAtlas or None, scaffold_path: str,
                   descriptor: ComplexDescriptor,
                   n_epochs: int, max_generations: int, generations_first_epoch: int, n_recomb: int, max_siblings: int,
                   n_top: int, n_clones: int, n_workers: int,
                   clustering_factor: float, score_threshold: float, distance_factor: float, orient_factor: float,
                   secor_factor: float, score_exponent: float,
                   clash_factor: float, clash_merge_base: float, delta_factor: float, n_delta: float,
                   std_dev_weight: float, n_quality_k: float,
                   deep_factor: float, random_mutations: bool, random_mutation_chance: float,
                   mutant_pattern: str, mutator_dumpfile: str, mutant_dir: str,
                   observer: Callable[[int, int, int], None] = None) -> GeneticMutator:
    """Performs an genetic mutation run based on a scaffold structure and a list of mutations to be applied to a
    dedicated host chain. If any of the mutations increases the total score, it will be taken into consideration for
    recombination with other successful mutations. This procedure is repeated until no more mutations are found or
    a given maximum number of generations has been reached. The output includes a ranking of all mutation sets attempted
    :param scorer: based on this instance of a subclass of BaseScorer the mutated structures are scored
    :param atlas: the Atlas to use for generating a new pmatrix in every epoch
    :param deep_atlas: the deep Atlas used to assist pmatrix calculation (optional)
    :param scaffold_path: the initial structure that does not contain any of the mutations
    :param descriptor: descriptor structure for ligand, binder, and coupled residues. From this descriptor, the
    protocol builds a new pmatrix and acomplex in every epoch
    :param n_epochs: the number of epochs (large iterations)
    :param max_generations: the maximum number of generations (small iterations) per epoch
    :param generations_first_epoch: the number of generations in first epoch. Equal max_generations if set 0
    :param n_recomb: the number of mutation sets to enqueue for recombination in each generation
    :param max_siblings: the maximum number of individuals with identical mutation sets that may coexist per epoch
    :param n_top: if the best n_top mutations do not change within a generation, the protocol terminates
    :param n_clones: this number of clones is created for every individual generated. Clones are then processed
    in parallel, each is carrying a unique individual ID. Original is also counted.
    :param n_workers: number of workers to use for parallel processing (-1 is for number of CPUs)
    :param clustering_factor: probability of distance clustering based mutation selection (default selection strategy)
    :param score_threshold: parameter for the scoring function: will be passed along while pmatrix calculation
    :param distance_factor: weight factor for distance coefficient of the scoring function. [1/Angstrom]
    :param orient_factor: weight factor for the primary orientation coefficient (C_alpha->C_beta) of the scoring
    function [1/radians]
    :param secor_factor: weight factor for the secondary orientation coefficient (C_alpha->C_O) of the scoring
    function [1/radians]
    :param score_exponent: the outcome of the scoring function is postprocessed by raising by this value
    :param clash_factor: probability to choose clash factoring for the selection of mutations in each selection round.
    :param clash_merge_base: Base of clash punishment for merging with normalized quality: n_quality*BASE^-clash.
    :param n_quality_k: Normalization constant for quality. Equals to quality value of 50%% in normalized state.
    :param n_delta: Normalization constant for delta. Equals to delta value of 50%% in normalized state.
    :param std_dev_weight: Base in weighing the standard deviation: BASE^-std_dev * clash.
    :param delta_factor: Correction factor for merging clash/quality with delta within calculation of mutation ranking:
    'n_quality*base^-clash + n_delta/FACTOR'.
    :param deep_factor: probability to involve a deep learning strategy for mutation selection
    :param random_mutations: whether to apply random mutations.
    :param random_mutation_chance: The chance of random mutations (over ones based on pmatrix).
    :param mutant_pattern: a file name pattern for mutated and scored structures. Must contain one $ symbol, which will
    be replaced by a unique identifier of the mutant
    :param mutator_dumpfile: if this file exists, the genetic mutator will be reconstructed from it. Furthermore, after
    every generation, the current progress is dumped into this file.
    :param mutant_dir: directory where mutant files are stored - relative to current directory.
    :param observer: callback function invoked after every recombination with current epoch, generation, recombination.
        None refers to return type.
    :return: a list of all individuals produced and scored in all epochs, ordered by score
    """
    def __dump_mutator(_mutator_to_dump: GeneticMutator) -> None:
        """Utility that pickles the mutator_to_dump.
        :param _mutator_to_dump: is pickled into mutator_dumpfile
        """
        if mutator_dumpfile is not None:
            with open(mutator_dumpfile, 'wb') as fo:
                pickle.dump(_mutator_to_dump, fo)

    def __load_mutator() -> GeneticMutator:
        """Utility that unpickles a mutator and returns it.
        Keep in mind that an epoch that has been discontinued will not be continued at the next generation
        after a restart. The next generation is epoch + 1 -> generation 1.
        :return: the instance unpickled from mutator_dumpfile
        """
        with open(mutator_dumpfile, 'rb') as fo:
            dumped_mutator = pickle.load(fo)
        return dumped_mutator

    def __mutate_and_score(_mut_indiv: Individual) -> None:
        """Nested worker function that applies the mutations described by one individual and scores the structure.
        :param _mut_indiv: the individual to mutate and score
        """
        job_start = timer()
        mutant_path = mutant_dir + ('' if mutant_dir.endswith('/') else '/') + \
                      mutant_pattern.replace("$", str(_mut_indiv.iid))
        log.info(f"\t\tCreating mutant {_mut_indiv.iid} based on template "
                 f"{_mut_indiv.template_structure}: ({_mut_indiv.mset})")
        job_parser = PDBParser(QUIET=True)
        template_structure: Structure = job_parser.get_structure(_mut_indiv.template_structure,
                                                                 _mut_indiv.template_structure)
        # TODO Make sure it works with more than one binder chain:
        mutate_mset(template_structure, descriptor.binder_info[0][0], _mut_indiv.absolute_mset())
        job_io = PDBIO()
        job_io.set_structure(template_structure)
        job_io.save(mutant_path)
        log.info(f"\t\tScoring mutant {_mut_indiv.iid}...")
        mutant_score = scorer.score(mutant_path, descriptor)
        if os.path.isfile(mutant_path):  # TODO make sure unrelaxed file is not needed any more
            os.remove(mutant_path)
        else:
            log.warning(f"Error: {mutant_path} file not found")
        job_end = timer()
        log.info(f"\t\tDone with mutant {_mut_indiv.iid}. Time elapsed: "
                 f"{str(timedelta(seconds=job_end - job_start)).split('.')[0]}. Total score: {mutant_score}")
        _mut_indiv.optimized_structure = mutant_score.optimized_structure
        _mut_indiv.fitness = mutant_score.score
        _mut_indiv.score_terms = mutant_score.properties

    if not os.path.exists(mutant_dir):
        os.mkdir(mutant_dir)
    if n_workers < 0:
        n_workers = os.cpu_count()
    if generations_first_epoch < 0:
        generations_first_epoch = max_generations
    log.info("Initialization of genetic protocol.")
    mutator: GeneticMutator = None
    try:
        mutator = __load_mutator()
        log.info(f"\tRecovered dumped mutator instance: {mutator_dumpfile}. Starting resume mode overriding parameters "
                 f"as requested.")
        mutator.atlas = atlas
        mutator.deep_atlas = deep_atlas
        mutator.descriptor = descriptor
        mutator.n_top = n_top
        mutator.clash_factor = clash_factor
        mutator.deep_factor = deep_factor
        mutator.random_mutations = random_mutations
        mutator.random_mutation_chance = random_mutation_chance
        mutator.max_siblings = max_siblings
        mutator.n_clones = n_clones
        mutator.clash_params = clash_merge_base, delta_factor, n_delta, n_quality_k, std_dev_weight
    except FileNotFoundError:
        log.info(f"\tNo such dumpfile found: {mutator_dumpfile}. Skipping resume mode.")
    except EOFError:
        log.error(f"\tError while unpickling dumped mutator {mutator_dumpfile}. Disabling dumping for this run!")
        mutator_dumpfile = None
    except TypeError:
        pass
    if mutator is None:
        log.info(f"\tCreating a new mutator instance based on parameters passed.")
        mutator = GeneticMutator(atlas, deep_atlas, descriptor, n_top, clustering_factor, clash_factor, deep_factor,
                                 random_mutations, random_mutation_chance, max_siblings, n_clones, clash_merge_base,
                                 delta_factor, n_delta,
                                 std_dev_weight, n_quality_k)
        log.info(f"\tApplying ligand mutations to scaffold structure {scaffold_path}...")
        mutate_scaffold(scaffold_path, descriptor)
    if deep_factor > 0.0:
        log.info("\tInitializing deep atlas...")
        mutator.initialize_deep_atlas()
    log.info("Initialization done.")
    if mutator.current_epoch() > 0 or mutator.generation > 0:
        log.info(f"Resuming dumped protocol {mutator_dumpfile} after {mutator.current_epoch()} epochs, "
                 f"{mutator.generation} generations.")
        unscored_indivs = [i for i in mutator.population if i.fitness is None]
        if len(unscored_indivs) > 0:
            log.info(f"There are {len(unscored_indivs)} individuals left unscored. Catching up on this first.")
            with ThreadPoolExecutor(max_workers=n_workers) as executor:
                for _ in executor.map(__mutate_and_score, unscored_indivs):
                    pass
    while mutator.current_epoch() < n_epochs:
        mutator.start_epoch(scaffold_path)
        log.info(f"Starting epoch {mutator.current_epoch()} of {n_epochs} "
                 f"using progenitor structure {mutator.current_progenitor().optimized_structure}.")
        log.info("Creating acomplex...")
        acomplexfile: str = f"epoch{mutator.current_epoch()}.acomplex"
        mutator.update_acomplex(acomplexfile)
        log.info(f"Stored acomplex under {acomplexfile}.")
        log.info("Creating pmatrix...")
        pmatrixfile: str = f"epoch{mutator.current_epoch()}.pmatrix"
        if os.path.isfile(pmatrixfile):
            mutator.pmatrix_by_epoch.append(read_matrix(pmatrixfile))
        else:
            if mutator.update_pmatrix(pmatrixfile,
                                      score_threshold=score_threshold,
                                      distance_factor=distance_factor,
                                      orient_factor=orient_factor,
                                      secor_factor=secor_factor,
                                      score_exponent=score_exponent) is None:
                log.warning(f"Generated pmatrix shows a total quality of 0.0. Aborting epoch {mutator.current_epoch()} "
                            f"after 0 generations.")
                break
        log.info(f"Stored pmatrix under {pmatrixfile}.")
        while mutator.generation < (generations_first_epoch if mutator.current_epoch() == 1 else max_generations):
            mutator.start_generation()
            log.info(f"\tGeneration {mutator.generation} of "
                     f"{generations_first_epoch if mutator.current_epoch() == 1 else max_generations} in epoch "
                     f"{mutator.current_epoch()} of {n_epochs}.")
            recomb_indivs: List[Individual] = []
            for recomb in range(0, n_recomb):
                log.info(f"\t\tRecombination {recomb + 1} of {n_recomb}.")
                fittest = mutator.sorted_population(True)
                template = mutator.current_progenitor()
                progen_indiv = mutator.create_progenitor_individual()
                n_rand = 0
                if mutator.current_iid == 0:
                    log.info("\t\t\tFilling up with scaffold.")
                    indiv = progen_indiv
                elif recomb >= len(fittest) or recomb >= min(n_recomb - 1, int(n_recomb * 0.75)):
                    log.info("\t\t\tFilling up with orphan child.")
                    n_rand, indiv = mutator.create_orphan_individual(progen_indiv, 0.75)
                else:
                    mother = fittest[recomb]
                    if random.random() < 0.25:
                        father = progen_indiv
                    else:
                        father_index = int(random.uniform(0, min(len(fittest),
                                                                 max(int(len(fittest) / 4), 4 * n_recomb))))
                        father = fittest[father_index]
                    log.info(f"\t\t\tMother (score={mother.fitness}): [{mother.mset}]")
                    log.info(f"\t\t\tFather (score={father.fitness}): [{father.mset}]")
                    n_rand, indiv = mutator.create_breeded_individual(mother, father, 0.25)
                    template = sorted([progen_indiv, mother, father],
                                      key=lambda i: float('inf') if i.fitness is None else i.fitness)[0]
                log.info(f"\t\t\tChild after {n_rand} random mutation(s): [{indiv.mset}]")
                indiv.template_structure = template.optimized_structure if template.optimized_structure is not None \
                    else template.template_structure
                mutator.insert_individual(indiv)
                recomb_indivs.append(indiv)
                for i in range(1, n_clones):
                    clone = copy(indiv)
                    mutator.insert_individual(clone)
                    log.info(f"\t\tCreated individual {clone.iid} as a clone of {indiv.iid}.")
                    recomb_indivs.append(clone)
            n = len(recomb_indivs)
            log.info(f"\tParallel processing of {n} mutants...")
            with ThreadPoolExecutor(max_workers=n_workers) as executor:
                i = 0
                for _ in executor.map(__mutate_and_score, recomb_indivs):
                    i += 1
                    if observer is not None:
                        observer(mutator.current_epoch(), mutator.generation, i)
            log.info("\tParallel processing done.")
            __dump_mutator(mutator)
            top_n_pop = mutator.sorted_population(True, mutator.n_top)
            log.info(f"\tTop {len(top_n_pop)} population after epoch {mutator.current_epoch()}, "
                     f"generation {mutator.generation}:")
            progen_mset = mutator.current_progenitor().absolute_mset()
            if len(progen_mset.mutations) > 0:
                log.info(f"\tAll individuals inherit the progenitor mutation set: [{progen_mset}]")
            for indiv in top_n_pop:
                log.info('\t\t{:10.4f} '.format(indiv.fitness) + '{:3} '.format(indiv.generation) +
                         '{:5} '.format(indiv.iid) + f"[{indiv.mset}]")
            if mutator.generation < (generations_first_epoch if mutator.current_epoch() == 1 else max_generations) \
                    and not mutator.has_improved_during_generation():
                log.warning(f"No improvement over previous top {n_top} ranking. Aborting epoch "
                            f"{mutator.current_epoch()} after {mutator.generation} generations.")
                break
        ov_top_n_pop = mutator.sorted_population(False, mutator.n_top)
        log.info(f"Overall top {len(ov_top_n_pop)} population after epoch {mutator.current_epoch()}:")
        for indiv in ov_top_n_pop:
            log.info(f"\t{indiv}")
    log.info("Genetic protocol done.")
    return mutator


class NegativeIndividual(Individual):
    """Instances of this class represent invididuals of the population for negative design."""

    def __init__(self, mset: MutationSet, nd_mset: MutationSet):
        """Creates a new individual based on its mutation set
        :param mset: the set of binder mutations this individual reflects
        :param nd_mset: the set of ligand mutations this individual reflects for negative design approach
        """
        super().__init__(mset)
        self.nd_mset = nd_mset
        self.nd_iid = None
        # other score sets are scores that result in repetition of scoring. Including fitness and score terms.
        self.other_score_sets: List[Tuple[float, Dict[str, float]]] = []

    def __str__(self) -> str:
        # return '{:10.4f} '.format(self.fitness) + '{:5} '.format(self.iid) + f"[{self.nd_mset}]" + " (" + \
        #        ' '.join([str(k) + ':' + '{:10.4f} '.format(v).strip() for k, v in self.score_terms.items()]) + ")"
        first = f"    [{self.nd_mset}] " + '{:10.4f} '.format(self.fitness) + " (" + \
               ' '.join([str(k) + ':' + '{:10.4f} '.format(v).strip() for k, v in self.score_terms.items()]) + ")"
        other = ""
        for score_set in self.other_score_sets:
            other += "\n" + " "*10 + '{:10.4f} '.format(score_set[0]) + " (" + \
                     ' '.join([str(k) + ':' + '{:10.4f} '.format(v).strip() for k, v in score_set[1].items()]) + ")"

        return first + other
